from importlib import resources
from logging import getLogger
from typing import Callable, Optional, cast

import dysts.flows
import dysts.systems
import numpy as np
import scipy
import sympy as sp
from dysts.base import DynSys

from ._dysts_to_sympy import dynsys_to_sympy
from ._plotting import plot_training_data
from ._typing import ExperimentResult, Float1D, SimProbData
from ._utils import _sympy_expr_to_feat_coeff

try:
    import jax

    from ._diffrax_solver import _gen_data_jax
except ImportError:
    pass

INTEGRATOR_KEYWORDS = {"rtol": 1e-12, "method": "LSODA", "atol": 1e-12}
MOD_LOG = getLogger(__name__)
LOCAL_DYNAMICS_PATH = resources.files("sindy_exp").joinpath("addl_attractors.json")

ODE_CLASSES = {
    klass.lower(): getattr(dysts.flows, klass)
    for klass in dysts.systems.get_attractor_list()
}


def gen_data(
    system: str,
    seed: int,
    n_trajectories: int = 1,
    ic_stdev: float = 3,
    noise_abs: Optional[float] = None,
    noise_rel: Optional[float] = None,
    dt: float = 0.01,
    t_end: float = 10,
    display: bool = False,
    array_namespace: str = "numpy",
) -> ExperimentResult[tuple[list[SimProbData], list[dict[sp.Expr, float]]]]:
    """Generate random training and test data

    An Experiment step according to the mitosis experiment runner.
    Note that test data has no noise.

    Arguments:
        system: the system to integrate
        seed (int): the random seed for number generation
        n_trajectories (int): number of trajectories of training data
        ic_stdev (float): standard deviation for generating initial conditions
        noise_abs (float): measurement noise standard deviation.
            Defaults to .1 if noise_rel is None.
        noise_rel (float): measurement noise-to-signal power ratio.
            Either noise_abs or noise_rel must be None.  Defaults to
            None.
        dt: time step for sample
        t_end: end time of simulation
        display: Whether to display graphics of generated data.

    Returns:
        dictionary of data and descriptive information
    """
    try:
        dyst_sys = ODE_CLASSES[system.lower()]()
    except KeyError as e:
        raise ValueError(
            f"Unknown system {system}.  Check {__name__}.ODE_CLASSES"
        ) from e
    input_features, sp_expr, sp_lambda = dynsys_to_sympy(dyst_sys)
    coeff_true = _sympy_expr_to_feat_coeff(sp_expr)
    rhsfunc = lambda t, X: dyst_sys.rhs(X, t)  # noqa: E731
    try:
        x0_center = cast(Float1D, dyst_sys.ic)
    except KeyError:
        x0_center = np.zeros((len(input_features)), dtype=np.float64)
    try:
        nonnegative = getattr(dyst_sys, "nonnegative", False)
    except KeyError:
        nonnegative = False
    if noise_abs is not None and noise_rel is not None:
        raise ValueError("Cannot specify both noise_abs and noise_rel")
    elif noise_abs is None and noise_rel is None:
        noise_abs = 0.1

    MOD_LOG.info(f"Generating {n_trajectories} trajectories of f{system}")
    prob_data_list: list[SimProbData] = []
    if array_namespace == "numpy":
        feature_names = [feat.name for feat in input_features]
        for _ in range(n_trajectories):
            seed += 1
            prob = _gen_data(
                rhsfunc,
                feature_names,
                seed,
                x0_center=x0_center,
                nonnegative=nonnegative,
                ic_stdev=ic_stdev,
                noise_abs=noise_abs,
                noise_rel=noise_rel,
                dt=dt,
                t_end=t_end,
            )
            prob_data_list.append(prob)
    elif array_namespace == "jax":
        try:
            jax  # type: ignore
        except NameError:
            raise ImportError(
                "jax data generation requested but diffrax or sympy2jax not"
                " installed"
            )
        this_seed = jax.random.PRNGKey(seed)  # type: ignore
        for _ in range(n_trajectories):
            this_seed, _ = jax.random.split(this_seed)  # type: ignore
            prob = _gen_data_jax(  # type: ignore
                sp_expr,
                input_features,
                this_seed,
                x0_center=x0_center,  # type: ignore # numpy->jax
                nonnegative=nonnegative,
                ic_stdev=ic_stdev,
                noise_abs=noise_abs,
                noise_rel=noise_rel,
                dt=dt,
                t_end=t_end,
            )
            prob_data_list.append(prob)
    else:
        raise ValueError(
            f"Unknown array_namespace {array_namespace}.  Must be 'numpy' or 'jax'"
        )
    if display and prob_data_list:
        sample = prob_data_list[0]
        assert sample.x_train_true is not None  # typing
        figs = plot_training_data(sample.t_train, sample.x_train, sample.x_train_true)
        figs[0].suptitle("Sample Trajectory")

    return {
        "data": (prob_data_list, coeff_true),
        "main": f"{n_trajectories} trajectories of {rhsfunc}",
        "metrics": {"rel_noise": noise_rel, "abs_noise": noise_abs},
    }


def _gen_data(
    rhs_func: Callable,
    input_features: list[str],
    seed: Optional[int],
    x0_center: Float1D,
    ic_stdev: float,
    noise_abs: Optional[float],
    noise_rel: Optional[float],
    nonnegative: bool,
    dt: float,
    t_end: float,
) -> SimProbData:
    rng = np.random.default_rng(seed)
    t_train = np.arange(0, t_end, dt)
    t_train_span = (t_train[0], t_train[-1])
    if nonnegative:
        shape = ((x0_center + 1) / ic_stdev) ** 2
        scale = ic_stdev**2 / (x0_center + 1)
        x0 = np.array([rng.gamma(k, theta) for k, theta in zip(shape, scale)])
    else:
        x0 = ic_stdev * rng.standard_normal(len(input_features)) + x0_center
    x_train = scipy.integrate.solve_ivp(
        rhs_func,
        t_train_span,
        x0,
        t_eval=t_train,
        **INTEGRATOR_KEYWORDS,
    ).y.T

    x_train_true = np.copy(x_train)
    x_train_true_dot = np.array([rhs_func(0, xi) for xi in x_train_true])
    if noise_rel is not None:
        noise_abs = np.sqrt(_signal_avg_power(x_train) * noise_rel)
    x_train = x_train + cast(float, noise_abs) * rng.standard_normal(x_train.shape)

    assert noise_abs is not None  # typing
    return SimProbData(
        t_train, x_train, input_features, x_train_true, x_train_true_dot, noise_abs
    )


def _max_amplitude(signal: np.ndarray, axis: int) -> float:
    return np.abs(scipy.fft.rfft(signal, axis=axis)[1:]).max() / np.sqrt(
        signal.shape[axis]
    )


def _signal_avg_power(signal: np.ndarray) -> float:
    return np.square(signal).mean()


def _register_dyst(klass: type[DynSys]) -> type[DynSys]:
    """Register a custom dysts DynSys class for use in sindy_exp data generation."""
    ODE_CLASSES[klass.__name__.lower()] = klass
    return klass


@_register_dyst
class LotkaVolterra(DynSys):
    """Lotka-Volterra (predator-prey) dynamical system."""

    nonnegative = True

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(  # type: ignore # dysts
        x, y, t: float, alpha, beta, gamma, delta
    ) -> np.ndarray:
        """LV dynamics

        Args:
            x: prey population
            y: predator population
            t: time (ignored, since autonomous)
            alpha: prey growth rate
            beta: predation rate
            delta: predator reproduction rate
            gamma: predator death rate
        """
        dxdt = alpha * x - beta * x * y
        dydt = delta * x * y - gamma * y
        return np.array([dxdt, dydt])


@_register_dyst
class Hopf(DynSys):
    """Hopf normal form dynamical system."""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(x, y, t: float, mu, omega, A) -> np.ndarray:  # type: ignore # dysts
        dxdt = mu * x - omega * y - A * (x**3 + x * y**2)
        dydt = omega * x + mu * y - A * (x**2 * y + y**3)
        return np.array([dxdt, dydt])


@_register_dyst
class SHO(DynSys):
    """Linear damped simple harmonic oscillator"""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(x, y, t: float, a, b, c, d) -> np.ndarray:  # type: ignore # dysts
        dxdt = a * x + b * y
        dydt = c * x + d * y
        return np.array([dxdt, dydt])


@_register_dyst
class CubicHO(DynSys):
    """Cubic damped harmonic oscillator."""

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(x, y, t: float, a, b, c, d) -> np.ndarray:  # type: ignore # dysts
        dxdt = a * x**3 + b * y**3
        dydt = c * x**3 + d * y**3
        return np.array([dxdt, dydt])


@_register_dyst
class VanDerPol(DynSys):
    """Van der Pol oscillator.

    dx/dt = y
    dy/dt = mu * (1 - x^2) * y - x
    """

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(x, x_dot, t: float, mu) -> np.ndarray:  # type: ignore # dysts
        dxdt = x_dot
        dx2dt2 = mu * (1 - x**2) * x_dot - x
        return np.array([dxdt, dx2dt2])


@_register_dyst
class Kinematics(DynSys):
    """One-dimensional kinematics with constant acceleration.

    dx/dt = v
    dv/dt = a
    """

    def __init__(self):
        super().__init__(metadata_path=LOCAL_DYNAMICS_PATH)  # type: ignore # dysts

    @staticmethod
    def _rhs(x, v, t: float, a) -> np.ndarray:  # type: ignore # dysts
        dxdt = v
        dvdt = a
        return np.array([dxdt, dvdt])



from logging import getLogger
from typing import Any, Literal, TypeVar, cast, overload

import matplotlib.pyplot as plt
import numpy as np
import sympy as sp

from ._plotting import (
    compare_coefficient_plots_from_dicts,
    plot_test_trajectory,
    plot_training_data,
)
from ._typing import (
    DynamicsTrialData,
    ExperimentResult,
    FullDynamicsTrialData,
    SimProbData,
    SINDyTrialUpdate,
    _BaseSINDy,
)
from ._utils import (
    _simulate_test_data,
    coeff_metrics,
    integration_metrics,
    unionize_coeff_dicts,
)

metric_ordering = {
    "coeff_precision": "max",
    "coeff_f1": "max",
    "coeff_recall": "max",
    "coeff_mae": "min",
    "coeff_mse": "min",
    "mse_plot": "min",
    "mae_plot": "min",
}


T = TypeVar("T", bound=int)
DType = TypeVar("DType", bound=np.dtype)
MOD_LOG = getLogger(__name__)


@overload
def fit_eval(
    data: tuple[list[SimProbData], list[dict[sp.Expr, float]]],
    model: _BaseSINDy,
    simulations: Literal[False],
    display: bool,
) -> ExperimentResult[DynamicsTrialData]: ...


@overload
def fit_eval(
    data: tuple[list[SimProbData], list[dict[sp.Expr, float]]],
    model: _BaseSINDy,
    simulations: Literal[True],
    display: bool,
) -> ExperimentResult[FullDynamicsTrialData]: ...


def fit_eval(
    data: tuple[list[SimProbData], list[dict[sp.Expr, float]]],
    model: Any,
    simulations: bool = True,
    display: bool = True,
) -> ExperimentResult:
    """Fit and evaluate a SINDy model on a set of trajectories.

    Args:
        data: Tuple of (trajectories, true_equations), where ``trajectories`` is
            a list of SimProbData objects and ``true_equations`` is a list of
            dictionaries mapping SymPy symbols to their true coefficients for
            each state coordinate.
        model: A SINDy-like model implementing the _BaseSINDy protocol.
        simulations: Whether to run forward simulations for evaluation.
        display: Whether to generate plots as part of evaluation.
    """
    model = cast(_BaseSINDy, model)
    for trajectory in data[0]:
        assert trajectory.x_train_true is not None
    trajectories, true_equations = data
    input_features = trajectories[0].input_features

    x_train = [traj.x_train for traj in trajectories]
    t_train = [traj.t_train for traj in trajectories]
    model.fit(x_train, t=t_train, feature_names=input_features)

    MOD_LOG.info(f"Fitting a model: {model}")
    coeff_true_dicts, coeff_est_dicts = unionize_coeff_dicts(model, true_equations)

    # Special workaround for pysindy's legacy WeakPDELibrary
    if hasattr(model.feature_library, "K"):
        # WeakPDE library fails to simulate, so insert nonweak library
        # to Pipeline and SINDy model.
        inner_lib = model.feature_library.function_library
        model.feature_library = inner_lib  # type: ignore  # TODO: Fix in pysindy

    # Special workaround for pysindy's bad (soon to be legacy) differentiation API
    if hasattr(model, "differentiation_method") and hasattr(
        model.differentiation_method, "smoothed_x_"
    ):
        smooth_x = []
        for traj in trajectories:
            model.differentiation_method(traj.x_train, t=traj.t_train)
            smooth_x.append(model.differentiation_method.smoothed_x_)
    else:  # using WeakPDELibrary
        smooth_x = x_train

    trial_data = DynamicsTrialData(
        trajectories=trajectories,
        true_equations=coeff_true_dicts,
        sindy_equations=coeff_est_dicts,
        model=model,
        input_features=input_features,
        smooth_train=smooth_x,
    )
    MOD_LOG.info(f"Evaluating a model: {model}")
    metrics = coeff_metrics(coeff_est_dicts, coeff_true_dicts)
    if simulations:
        sims: list[SINDyTrialUpdate] = []
        integration_metric_list: list[dict[str, float | np.floating]] = []
        for traj in trajectories:
            sim = _simulate_test_data(model, traj.t_train, traj.x_train_true)
            sims.append(sim)
            integration_metric_list.append(
                integration_metrics(
                    model,
                    traj.x_train_true,
                    traj.t_train,
                    traj.x_train_true_dot,
                )
            )

        agg_integration_metrics: dict[str, float] = {}
        for key in integration_metric_list[0].keys():
            values = cast(list[float], [m[key] for m in integration_metric_list])
            agg_integration_metrics[key] = float(np.mean(values))
        metrics.update(agg_integration_metrics)

        trial_data = FullDynamicsTrialData(sims=sims, **trial_data.__dict__)
    if display:
        plot_ode_panel(trial_data)
        for i, traj in enumerate(trajectories):
            fig_composite, fig_by_coord_1d = plot_training_data(
                traj.t_train,
                traj.x_train,
                traj.x_train_true,
                x_smooth=smooth_x[i],
                coord_names=input_features,
            )
            if simulations:
                # Overlay test trajectory time series on the coordinate-wise figure
                plot_test_trajectory(
                    traj.x_train_true,
                    sims[i].x_sim,
                    traj.t_train,
                    sims[i].t_sim,
                    figs=(fig_composite, fig_by_coord_1d),
                    coord_names=input_features,
                )

    return {"metrics": metrics, "data": trial_data, "main": metrics["main"]}


def plot_ode_panel(trial_data: DynamicsTrialData):
    trial_data.model.print()
    compare_coefficient_plots_from_dicts(
        trial_data.sindy_equations,
        trial_data.true_equations,
        input_features=[_texify(feat) for feat in trial_data.input_features],
    )
    plt.show()


def _texify(input: str) -> str:
    if input[0] != "$":
        input = "$" + input
    if input[-1] != "$":
        input = input + "$"
    return input
